""" Functions for performing statistical inference using surrogate maps. """

import numpy as np

__all__ = ['pearsonr', 'pairwise_r', 'nonparp']


def pearsonr(X, Y):
    """
    Multi-dimensional Pearson correlation between rows of `X` and `Y`.

    Parameters
    ----------
    X : (N,P) np.ndarray
    Y : (M,P) np.ndarray

    Returns
    -------
    (N,M) np.ndarray

    Raises
    ------
    TypeError : `x` or `y` is not array_like
    ValueError : `x` and `y` are not same size along second axis

    """
    if not isinstance(X, np.ndarray) or not isinstance(Y, np.ndarray):
        raise TypeError('X and Y must be numpy arrays')

    if X.ndim == 1:
        X = X.reshape(1, -1)
    if Y.ndim == 1:
        Y = Y.reshape(1, -1)

    n = X.shape[1]
    if n != Y.shape[1]:
        raise ValueError('X and Y must be same size along axis=1')

    mu_x = X.mean(axis=1)
    mu_y = Y.mean(axis=1)

    s_x = X.std(axis=1, ddof=n - 1)
    s_y = Y.std(axis=1, ddof=n - 1)
    cov = np.dot(X, Y.T) - n * np.dot(
        mu_x[:, np.newaxis], mu_y[np.newaxis, :])
    return cov / np.dot(s_x[:, np.newaxis], s_y[np.newaxis, :])


def pairwise_r(X, flatten=False):
    """
    Compute pairwise Pearson correlations between rows of `X`.

    Parameters
    ----------
    X : (N,M) np.ndarray
    flatten : bool, default False
        If True, return flattened upper triangular elements of corr. matrix

    Returns
    -------
    (N*(N-1)/2,) or (N,N) np.ndarray
        Pearson correlation coefficients

    """
    rp = pearsonr(X, X)
    if not flatten:
        return rp
    triu_inds = np.triu_indices_from(rp, k=1)
    return rp[triu_inds].flatten()


def nonparp(stat, dist):
    """
    Compute two-sided non-parametric p-value.

    Compute the fraction of elements in `dist` which are more extreme than
    `stat`.

    Parameters
    ----------
    stat : float
        Test statistic
    dist : (N,) np.ndarray
        Null distribution for test statistic

    Returns
    -------
    float
        Fraction of elements in `dist` which are more extreme than `stat`

    """
    n = float(len(dist))
    return np.sum(np.abs(dist) > abs(stat)) / n



""" Evaluation metrics for randomly generated surrogate maps. """

from brainsmash.mapgen.base import Base
from brainsmash.mapgen.sampled import Sampled
from brainsmash.utils.dataio import dataio
import matplotlib.pyplot as plt
import numpy as np

__all__ = ['base_fit', 'sampled_fit']


def base_fit(x, D, nsurr=100, **params):
    """
    Evaluate variogram fits for Base class.

    Parameters
    ----------
    x : (N,) np.ndarray or filename
        Target brain map
    D : (N,N) np.ndarray or filename
        Pairwise distance matrix between regions in `x`
    nsurr : int, default 100
        Number of simulated surrogate maps from which to compute variograms
    params
        Keyword arguments for :class:`brainsmash.mapgen.base.Base`

    Returns
    -------
    None

    Notes
    -----
    Generates and shows a matplotlib plot instance illustrating the fit of
    the surrogates' variograms to the target map's variogram.

    """

    x = dataio(x)
    d = dataio(D)

    # Instantiate surrogate map generator
    generator = Base(x=x, D=d, **params)

    # Simulate surrogate maps
    surrogate_maps = generator(n=nsurr)

    # Compute empirical variogram
    v = generator.compute_variogram(x)
    emp_var, u0 = generator.smooth_variogram(v, return_h=True)

    # Compute surrogate map variograms
    surr_var = np.empty((nsurr, generator.nh))
    for i in range(nsurr):
        v_null = generator.compute_variogram(surrogate_maps[i])
        surr_var[i] = generator.smooth_variogram(v_null, return_h=False)

    # # Create plot for visual comparison

    # Plot empirical variogram
    fig = plt.figure(figsize=(3, 3))
    ax = fig.add_axes([0.12, 0.15, 0.8, 0.77])
    ax.autoscale(axis='y', tight=True)

    ax.scatter(u0, emp_var, s=20, facecolor='none', edgecolor='k',
               marker='o', lw=1, label='Empirical')

    # Plot surrogate maps' variograms
    mu = surr_var.mean(axis=0)
    sigma = surr_var.std(axis=0)
    ax.fill_between(u0, mu-sigma, mu+sigma, facecolor='#377eb8',
                    edgecolor='none', alpha=0.3)
    ax.plot(u0, mu, color='#377eb8', label='SA-preserving', lw=1)

    # Make plot nice
    leg = ax.legend(loc=0)
    leg.get_frame().set_linewidth(0.0)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.setp(ax.get_yticklabels(), visible=False)
    plt.setp(ax.get_yticklines(), visible=False)
    plt.setp(ax.get_xticklabels(), visible=False)
    plt.setp(ax.get_xticklines(), visible=False)
    ax.set_xlabel("Spatial separation\ndistance")
    ax.set_ylabel("Variance")

    plt.show()


def sampled_fit(x, D, index, nsurr=10, **params):
    """
    Evaluate variogram fits for Sampled class.

    Parameters
    ----------
    x : (N,) np.ndarray
        Target brain map
    D : (N,N) np.ndarray or np.memmap
        Pairwise distance matrix between elements of `x`
    index : (N,N) np.ndarray or np.memmap
        See :class:`brainsmash.mapgen.sampled.Sampled`
    nsurr : int, default 10
        Number of simulated surrogate maps from which to compute variograms
    params
        Keyword arguments for :class:`brainsmash.mapgen.sampled.Sampled`

    Returns
    -------
    None

    Notes
    -----
    Generates and shows a matplotlib plot instance illustrating the fit of
    the surrogates' variograms to the target map's variogram.

    """

    # Instantiate surrogate map generator
    generator = Sampled(x=x, D=D, index=index, **params)

    # Simulate surrogate maps
    surrogate_maps = generator(n=nsurr)

    # Compute target & surrogate map variograms
    surr_var = np.empty((nsurr, generator.nh))
    emp_var_samples = np.empty((nsurr, generator.nh))
    u0_samples = np.empty((nsurr, generator.nh))
    for i in range(nsurr):
        idx = generator.sample()  # Randomly sample a subset of brain areas
        v = generator.compute_variogram(generator.x, idx)
        u = generator.D[idx, :]
        umax = np.percentile(u, generator.pv)
        uidx = np.where(u < umax)
        emp_var_i, u0i = generator.smooth_variogram(
            u=u[uidx], v=v[uidx], return_h=True)
        emp_var_samples[i], u0_samples[i] = emp_var_i, u0i
        # Surrogate
        v_null = generator.compute_variogram(surrogate_maps[i], idx)
        surr_var[i] = generator.smooth_variogram(
            u=u[uidx], v=v_null[uidx], return_h=False)

    # # Create plot for visual comparison
    u0 = u0_samples.mean(axis=0)
    emp_var = emp_var_samples.mean(axis=0)

    # Plot target variogram
    fig = plt.figure(figsize=(3, 3))
    ax = fig.add_axes([0.12, 0.15, 0.8, 0.77])
    ax.scatter(u0, emp_var, s=20, facecolor='none', edgecolor='k',
               marker='o', lw=1, label='Empirical')

    # Plot surrogate maps' variograms
    mu = surr_var.mean(axis=0)
    sigma = surr_var.std(axis=0)
    ax.fill_between(u0, mu-sigma, mu+sigma, facecolor='#377eb8',
                    edgecolor='none', alpha=0.3)
    ax.plot(u0, mu, color='#377eb8', label='SA-preserving', lw=1)

    # Make plot nice
    leg = ax.legend(loc=0)
    leg.get_frame().set_linewidth(0.0)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    plt.setp(ax.get_yticklabels(), visible=False)
    plt.setp(ax.get_yticklines(), visible=False)
    plt.setp(ax.get_xticklabels(), visible=False)
    plt.setp(ax.get_xticklines(), visible=False)
    ax.set_xlabel("Spatial separation\ndistance")
    ax.set_ylabel("Variance")
    plt.show()



from .checks import check_sampled, check_file_exists, check_distmat
from .checks import check_map, check_deltas, check_pv, check_outfile
from .checks import check_extensions, count_lines, is_string_like, stripext
from .dataio import dataio



""" Routines for constructing distance matrices from neuroimaging files. """

from ..utils.checks import *
from .io import check_surface, check_image_file
from ..utils.dataio import load, export_cifti_mapping, dataio
from ..config import parcel_labels_lr
from .surf import make_surf_graph
from scipy.spatial.distance import cdist
from scipy import ndimage, sparse
from tempfile import gettempdir, NamedTemporaryFile
from os import path
from os import system
import numpy as np
import nibabel as nib
import numpy.lib.format
# from warnings import warn
# from ..utils.dataio import dataio
# from os import remove
from joblib import Parallel, delayed

__all__ = ['cortex', 'subcortex', 'parcellate', 'volume']


def cortex(surface, outfile, euclid=False, dlabel=None, medial=None,
           use_wb=True, unassigned_value=0, verbose=True, n_jobs=None):
    """
    Calculates surface distances for `surface` mesh and saves to `outfile`.

    Parameters
    ----------
    surface : str or os.PathLike
        Path to surface file on which to calculate distance
    outfile : str or os.PathLike
        Path to which generated distance matrix should be saved
    euclid : bool, optional, default False
        Whether to compute Euclidean distance instead of surface distance
    dlabel : str or os.PathLike, optional, default None
        Path to file with parcel labels for provided `surf`. If provided,
        calculate and save parcel-parcel distances instead of vertex distances.
    medial : str or os.PathLike, optional, default None
        Path to file containing labels for vertices corresponding to medial
        wall. If provided and `use_wb=False`, will disallow calculation of
        surface distance along the medial wall.
    use_wb : bool, optional, default True
        Whether to use calls to `wb_command -surface-geodesic-distance` for
        computation of the surface distance matrix; this will involve
        significant disk I/O. If False, all computations will be done in memory
        using the `scipy.sparse.csgraph.dijkstra` function.
    unassigned_value : int, default 0
        Label value which indicates that a vertex/voxel is not assigned to
        any parcel. This label is excluded from the output. 0 is the default
        value used by Connectome Workbench, e.g. for ``-cifti-parcellate``.
    verbose : bool, optional, default True
        Whether to print status updates while distances are calculated.
    n_jobs : int, default None
        The number of parallel jobs to run for distance calculation. None means
        1 unless in a ``joblib.parallel_backend`` context. -1 means using all
        processors.

    Returns
    -------
    distance : str
        Path to generated `outfile`

    Notes
    -----
    The surface distance matrix computed with `use_wb=False` will have slightly
    lower values than when `use_wb=True` due to known estimation errors. These
    will be fixed at a later date. By default, `use_wb=True` for backwards-
    compatibility but this will be changed in a future update.

    Raises
    ------
    ValueError : inconsistent # of vertices in label, mask, and/or surface file

    """

    n_vert = len(load(surface))

    # get data from dlabel / medial wall files if provided
    labels, mask = None, np.zeros(n_vert, dtype=bool)
    if dlabel is not None:
        labels = check_image_file(dlabel)
        if len(labels) != n_vert:
            raise ValueError('Provided `dlabel` file does not contain same '
                             'number of vertices as provided `surface`')
    if medial is not None:
        mask = np.asarray(check_image_file(medial), dtype=bool)
        if len(mask) != n_vert:
            raise ValueError('Provided `medial` file does not contain same '
                             'number of vertices as provided `surface`')

    # define which function we'll be using to calculate the distances
    if euclid:
        func = _get_euclid_distance
        graph = check_surface(surface)  # vertex coordinates
    else:
        if use_wb:
            func = _get_workbench_distance
            graph = surface
        else:
            func = _get_graph_distance
            vert, faces = [darray.data for darray in nib.load(surface).darrays]
            graph = make_surf_graph(vert, faces, mask=mask)

    # if we want the vertex-vertex distance matrix we'll stream it to disk to
    # save on memory, a la `_geodesic()` or `_euclid()`
    # NOTE: streaming to disk takes a lot more _time_ than storing in memory
    if labels is None:
        with open(outfile, 'w') as dest:
            for n in range(n_vert):
                if verbose and n % 1000 == 0:
                    print('Running vertex {} of {}'.format(n, n_vert))
                np.savetxt(dest, func(n, graph))
    # we can store the temporary n_vert x label matrix in memory; running this
    # is much faster than trying to read through the giant vertex-vertex
    # distance matrix file
    else:
        # depends on size of parcellation, but assuming even a liberal 1000
        # parcel atlas this will be ~250 MB in-memory for the default fslr32k
        # resolution
        unique_parcels = np.unique(labels)
        par, func = Parallel(n_jobs=n_jobs), delayed(func)
        dist = np.row_stack(par(func(n, graph, labels) for n in range(n_vert)))
        # average rows (vertices) into parcels; columns are already parcels
        dist = np.row_stack([
            dist[labels == lab].mean(axis=0) for lab in unique_parcels])
        dist[np.diag_indices_from(dist)] = 0
        # NOTE: if `medial` is supplied and any of the parcel labels correspond
        # to the medial wall then those parcel-parcel distances will be `inf`!

        # remove unassigned parcel
        if unassigned_value in unique_parcels:
            idx = list(unique_parcels).index(unassigned_value)
            dist = np.delete(dist, idx, axis=0)
            dist = np.delete(dist, idx, axis=1)

        np.savetxt(outfile, dist)

    return outfile


# def cortex(surface, outfile, euclid=False):
#     """
#     Compute distance matrix for a cortical hemisphere.
#
#     Parameters
#     ----------
#     surface : filename
#         Path to a surface GIFTI (.surf.gii) from which to compute distances
#     outfile : filename
#         Path to output file
#     euclid : bool, default False
#         If True, compute Euclidean distances; if False, compute geodesic dist
#
#     Returns
#     -------
#     filename : str
#         Path to output distance matrix file
#
#     """
#
#     check_outfile(outfile)
#
#     # Strip file extensions and define output text file
#     outfile = stripext(outfile)
#     dist_file = outfile + '.txt'
#
#     # Load surface file
#     coords = check_surface(surface)
#
#     if euclid:  # Pairwise Euclidean distance matrix
#         of = _euclidean(dist_file=dist_file, coords=coords)
#     else:  # Pairwise geodesic distance matrix
#         of = _geodesic(
#             surface=surface, dist_file=dist_file, coords=coords)
#     return of


def subcortex(fout, image_file=None, dlabel=None, unassigned_value=0,
              verbose=True):
    """
    Compute inter-voxel Euclidean distance matrix.

    Parameters
    ----------
    fout : str or os.Pathlike
        Path to output text file
    image_file : str or os.Pathlike or None, default None
        Path to a CIFTI-2 format neuroimaging file (eg .dscalar.nii). MNI
        coordinates for each subcortical voxel are read from this file's
        metadata. If None, uses dlabel file defined in ``brainsmash.config.py``.
    dlabel : str or os.PathLike, optional, default None
        Path to file with parcel labels for provided `surf`. If provided,
        calculate and save parcel-parcel distances instead of vertex distances.
    unassigned_value : int, default 0
        Label value which indicates that a vertex/voxel is not assigned to
        any parcel. This label is excluded from the output. 0 is the default
        value used by Connectome Workbench, e.g. for ``-cifti-parcellate``.
    verbose : bool, optional, default True
        Whether to print status updates while distances are calculated.

    Returns
    -------
    filename : str
        Path to output text file containing pairwise Euclidean distances

    Notes
    -----
    Voxel indices are used as a proxy for physical distance, since the two are
    related by a simple linear scaling. Note that this assumes voxels are
    cubic, i.e., that the scaling is equivalent along the x, y, and z dimension.

    Raises
    ------
    ValueError : `image_file` header does not contain volume information
    IndexError : Inconsistent number of elements in `image_file` and `dlabel`

    """
    # TODO add more robust error handling

    check_outfile(fout)

    # Strip file extensions and define output text file
    fout = stripext(fout)
    dist_file = fout + '.txt'

    # Load CIFTI mapping  (i.e., map from scalar index to 3-D MNI indices)
    if image_file is None:
        image_file = parcel_labels_lr
    maps = export_cifti_mapping(image_file)
    if "volume" not in maps.keys():
        e = "Subcortical information was not found in {}".format(image_file)
        raise ValueError(e)
    coords = maps['volume'].drop("structure", axis=1).values
    # outfile = _euclidean(dist_file=dist_file, coords=coords)
    n_vert = coords.shape[0]

    # Get data from dlabel file if provided
    labels, mask = None, np.zeros(n_vert, dtype=bool)
    if dlabel is not None:
        all_labels = check_image_file(dlabel)
        volume_indices = maps['volume'].index.values
        try:
            labels = all_labels[volume_indices]
        except IndexError:
            raise IndexError(
                'Volumetric CIFTI indices obtained from `image_file` could not '
                'be indexed from the provided `dlabel` file.')

    func = _get_euclid_distance

    # If we want the vertex-vertex distance matrix we'll stream it to disk to
    # save on memory.
    # NOTE: streaming to disk takes a lot more _time_ than storing in memory
    if labels is None:
        with open(dist_file, 'w') as dest:
            for n in range(n_vert):
                if verbose and n % 1000 == 0:
                    print('Running vertex {} of {}'.format(n, n_vert))
                np.savetxt(dest, func(n, coords))
    # We can store the temporary n_vert x label matrix in memory; running this
    # is much faster than trying to read through the giant vertex-vertex
    # distance matrix file
    else:
        # depends on size of parcellation, but assuming even a liberal 1000
        # parcel atlas this will be ~250 MB in-memory for the default fslr32k
        # resolution
        unique_parcels = np.unique(labels)
        dist = np.zeros((n_vert, unique_parcels.size), dtype='float32')
        # NOTE: because this is being done in-memory it could be multiprocessed
        # for additional speed-ups, if desired!
        for n in range(n_vert):
            if verbose and n % 1000 == 0:
                print('Running vertex {} of {}'.format(n, n_vert))
            dist[n] = func(n, coords, labels)
        # average rows (vertices) into parcels; columns are already parcels
        dist = np.row_stack([
            dist[labels == lab].mean(axis=0) for lab in unique_parcels])
        dist[np.diag_indices_from(dist)] = 0

        # remove unassigned parcel
        if unassigned_value in unique_parcels:
            idx = list(unique_parcels).index(unassigned_value)
            dist = np.delete(dist, idx, axis=0)
            dist = np.delete(dist, idx, axis=1)

        np.savetxt(dist_file, dist)

    return dist_file


def parcellate(infile, dlabel_file, outfile, delimiter=' ', unassigned_value=0):
    """
    Parcellate a dense distance matrix.

    Parameters
    ----------
    infile : filename
        Path to `delimiter`-separated distance matrix file
    dlabel_file : filename
        Path to parcellation file  (.dlabel.nii)
    outfile : filename
        Path to output text file (to be created)
    delimiter : str, default ' '
        Delimiter between elements in `infile`
    unassigned_value : int, default 0
        Label value which indicates that a vertex/voxel is not assigned to
        any parcel. This label is excluded from the output. 0 is the default
        value used by Connectome Workbench, e.g. for ``-cifti-parcellate``.

    Returns
    -------
    filename : str
        Path to output parcellated distance matrix file

    Notes
    -----
    For two parcels A and B, the inter-parcel distance is defined as the mean
    distance between area i in parcel A and area j in parcel B, for all i,j.

    Inputs `infile` and `dlabel_file` should include the same anatomical
    structures, e.g. the left cortical hemisphere, and should have the same
    number of elements. If you need to isolate one anatomical structure from
    `dlabel_file`, see the following Workbench function:
    https://www.humanconnectome.org/software/workbench-command/-cifti-separate

    Raises
    ------
    ValueError : `infile` and `dlabel_file` have inconsistent sizes

    """

    print("\nComputing parcellated distance matrix\n")
    m = "For a 32k-vertex cortical hemisphere, this takes about 30 mins "
    m += "for the HCP MMP parcellation. For subcortex, this takes about an hour"
    m += " for the CAB-NP parcellation."
    print(m)

    check_outfile(outfile)

    # Strip file extensions and define output text file
    fout = stripext(outfile)
    dist_file = fout + '.txt'

    # Load parcel labels
    labels = check_image_file(dlabel_file)

    with open(infile, 'r') as fp:

        # Compare number of elements in distance matrix to dlabel file
        nrows = 1
        ncols = len(fp.readline().split(delimiter))
        for l in fp:
            if l.rstrip():
                nrows += 1
        if not (labels.size == nrows == ncols):
            e = "Files must contain same number of areas\n"
            e += "{} areas in {}\n".format(labels.size, dlabel_file)
            e += "{} rows and {} cols in {}".format(nrows, ncols, infile)
            raise ValueError(e)
        fp.seek(0)  # return to beginning of file

        # Skip unassigned parcel label
        unique_labels = np.unique(labels)
        nparcels = unique_labels.size
        if unassigned_value in unique_labels:
            unique_labels = unique_labels[unique_labels != unassigned_value]
            nparcels -= 1

        # Create vertex-level mask for each unique cortical parcel
        parcel_vertex_mask = {l: labels == l for l in unique_labels}

        # Loop over pairwise parcels at the level of surface vertices
        distance_matrix = np.zeros((nparcels, nparcels))

        for i, li in enumerate(unique_labels[:-1]):

            # Labels of parcels for which to compute mean geodesic distance
            labels_lj = unique_labels[i+1:]

            # Initialize lists in which to store pairwise vertex-level distances
            parcel_distances = {lj: list() for lj in labels_lj}

            # Loop over vertices with parcel label i
            li_vertices = np.where(parcel_vertex_mask[li])[0]

            fp.seek(0)
            for vi, l in enumerate(fp):
                if vi in li_vertices:
                    # Load distance from vertex vi to every other vertex
                    d = np.array(l.split(delimiter), dtype=np.float32)
                    # Store dists from vertex vi to vertices in parcel j
                    for lj in labels_lj:
                        vi_lj_distances = d[parcel_vertex_mask[lj]]
                        parcel_distances[lj].append(vi_lj_distances)

            # Compute average geodesic distances
            for j, lj in enumerate(labels_lj):
                mean_distance = np.mean(parcel_distances[lj])
                distance_matrix[i, i + j + 1] = mean_distance

            print("# Parcel label %s complete." % str(li))

        # Make final matrix symmetric
        i, j = np.triu_indices(nparcels, k=1)
        distance_matrix[j, i] = distance_matrix[i, j]

        # Write to file
        np.savetxt(fname=dist_file, X=distance_matrix)
        check_file_exists(dist_file)
        return dist_file


def volume(coord_file, outdir, chunk_size=1000):
    """
    Generate distance-matrix-related memory-mapped files for volumetric data.

    Parameters
    ----------
    coord_file : str or os.PathLike
        Path to text file in which rows correspond to voxels of interest, and
        *three* columns correspond to voxels' coordinates
    outdir : str or os.PathLike
        Path to the directory where outputs will be saved
    chunk_size : int, default 1000
        The number of voxels to process per chunk. For N voxels, this will
        impose a memory burden of N*`chunk_size` per iteration (in contrast to
        a memory burden of N*N for a single iteration, in the absence of
        chunking).

    Returns
    -------
    dict
        Keys are 'D' and 'index'; values are absolute paths to the
        corresponding files on disk. These files are used as inputs to
        `brainsmash.mapgen.sampled.Sampled`.

    Raises
    ------
    IOError : `outdir` doesn't exist
    ValueError : `coord_file` does not contain three columns

    Notes
    -----
    This function computes 3D Euclidean distance between all pairs of voxels
    whose 3D coordinates are provided in `coord_file`. The distance matrix is
    not saved in its raw, symmetric form, but rather as a pair of memory-mapped
    arrays which are needed to create an instance of the
    `brainsmash.mapgen.sampled.Sampled` class. See
    `brainsmash.mapgen.memmap.txt2memmap` for more details. Note that the input
    file should only contain brain voxels --- i.e., voxels in your brain map
    of interest. Each row of `coord_file`, which indicates the coordinates of
    a voxel, should therefore correspond to one brain map value.

    """
    if not path.exists(outdir):
        raise IOError("Output directory does not exist: {}".format(outdir))

    # Load data
    X = dataio(coord_file)
    if X.ndim != 2 or X.shape[1] != 3:
        e = f"expected N rows by 3 columns, instead got shape {X.shape}"
        raise ValueError(e)

    n = X.shape[0]
    print(f"loading voxels coordinates from {coord_file}")
    print(f"file contains {n} voxels")
    print(f"saving memory-mapped distance matrix files to {outdir}")

    # Open memory-mapped arrays
    npydfile = path.join(outdir, "distmat.npy")
    npyifile = path.join(outdir, "index.npy")
    fpd = numpy.lib.format.open_memmap(
        npydfile, mode='w+', dtype=np.float32, shape=(n, n))
    fpi = numpy.lib.format.open_memmap(
        npyifile, mode='w+', dtype=np.int32, shape=(n, n))

    i = 0
    while i < n:
        j = min(i + chunk_size, n)
        d = cdist(X[i:j], X)  # compute 3D euclidean distance
        sort_idx = np.argsort(d, axis=1)  # sort each row
        fpd[i:j] = d[np.arange(j-i)[:, np.newaxis], sort_idx]
        fpi[i:j] = sort_idx  # indices used to sort each row
        i += chunk_size

    del fpd  # Flush memory changes to disk
    del fpi

    return {'D': npydfile, 'index': npyifile}  # Return filenames

# def euclidean(coords, fout, chunk_size=10000):
#     """
#     Generate a three-dimensional Euclidean distance matrix and write to file.
#
#     Parameters
#     ----------
#     coords : (N,3) np.ndarray or str or os.Pathlike
#         3-D coordinates for N brain regions (e.g. voxels, parcels), provided
#         either as a numpy array or as a filename
#     fout : filename
#         Outputs are written to this file; include .txt extension!
#     chunk_size : float
#         The distance matrix is written to file `chunk_size` rows at
#         a time, assuming N is larger than `chunk_size`.
#
#     Returns
#     -------
#     filename : str
#         Output text file
#
#     Raises
#     ------
#     TypeError : coordinates not a numpy array
#     ValueError : coordinates not a two-dimensional (N,3) matrix
#
#     """
#
#     X = dataio(coords)
#     if not isinstance(X, np.ndarray):
#         e = "Expected array-like, got type {}".format(type(X))
#         raise TypeError(e)
#     if X.ndim != 2 or X.shape[1] != 3:
#         e = "Expected (N,3)-shaped coordinates, got shape {}".format(X.shape)
#         raise ValueError(e)
#     if path.exists(fout):
#         warn("Output file already exists, will be overwritten")
#         if input("continue? [y/n] ") == "y":
#             remove(fout)
#         else:
#             print("terminating")
#             return
#
#     n = X.shape[0]
#     with open(fout, 'ab') as fp:
#         i = 0
#         while i < n:
#             j = min(i+chunk_size, n)
#             print(i, j)
#             Y = X[i:i+j]
#             d = cdist(Y, X)
#             np.save(fp, d)
#             i += chunk_size
#     check_file_exists(f=fout)
#     return fout


def _geodesic(surface, dist_file, coords):
    """
    Compute pairwise geodesic distance between rows of `coords`. Write results
    to `dist_file`.

    Parameters
    ----------
    surface : filename
        Path to a surface GIFTI file (.surf.gii) from which to compute distances
    dist_file : filename
        Path to output file, with .txt extension
    coords : (N,3) np.ndarray
        MNI coordinates for N voxels/vertices

    Returns
    -------
    filename : str
        Path to output distance matrix file

    Notes
    -----
    This function uses command-line utilities included in Connectome Workbench.

    """
    nvert = coords.shape[0]

    print("\nComputing geodesic distance matrix\n")
    m = "For a 32k-vertex cortical hemisphere, this may take up to two hours."
    print(m)

    # Files produced at runtime by Workbench commands
    temp = gettempdir()
    coord_file = path.join(temp, "coords.func.gii")
    distance_metric_file = path.join(temp, "dist.func.gii")

    # Create a metric file containing the coordinates of each surface vertex
    cmd = 'wb_command -surface-coordinates-to-metric "{0:s}" "{1:s}"'
    system(cmd.format(surface, coord_file))

    with open(dist_file, 'w') as f:
        for ii in np.arange(coords.shape[0]):
            cmd = 'wb_command -surface-geodesic-distance "{0}" {1} "{2}" '
            system(cmd.format(surface, ii, distance_metric_file))
            distance_from_iv = load(distance_metric_file)
            line = " ".join([str(dij) for dij in distance_from_iv])
            f.write(line + "\n")
            if not (ii % 1000):
                print("Vertex {} of {} complete.".format(ii+1, nvert))
    check_file_exists(f=dist_file)
    return dist_file


def _get_workbench_distance(vertex, surf, labels=None):
    """
    Gets surface distance of `vertex` to all other vertices in `surf`.

    Parameters
    ----------
    vertex : int
        Index of vertex for which to calculate surface distance
    surf : str or os.PathLike
        Path to surface file on which to calculate distance
    labels : array_like, optional (default None)
        Labels indicating parcel to which each vertex belongs. If provided,
        distances will be averaged within distinct labels.

    Returns
    -------
    dist : (N,) numpy.ndarray
        Distance of `vertex` to all other vertices in `graph` (or to all
        parcels in `labels`, if provided)

    """

    distcmd = 'wb_command -surface-geodesic-distance {surf} {vertex} {out}'

    # run the geodesic distance command with wb_command
    with NamedTemporaryFile(suffix='.func.gii') as out:
        system(distcmd.format(surf=surf, vertex=vertex, out=out.name))
        dist = load(out.name)

    return _get_parcel_distance(vertex, dist, labels)


def _get_graph_distance(vertex, graph, labels=None):
    """
    Gets surface distance of `vertex` to all other vertices in `graph`

    Parameters
    ----------
    vertex : int
        Index of vertex for which to calculate surface distance
    graph : array_like
        Graph along which to calculate shortest path distances
    labels : array_like, optional
        Labels indicating parcel to which each vertex belongs. If provided,
        distances will be averaged within unique labels

    Returns
    -------
    dist : (N,) numpy.ndarray
        Distance of `vertex` to all other vertices in `graph` (or to all
        parcels in `labels`, if provided)

    Notes
    -----
    Distances are computed using Dijkstra's algorithm.

    """

    # this involves an up-cast to float64; will produce some numerical rounding
    # discrepancies here when compared to the wb_command subprocess call
    dist = sparse.csgraph.dijkstra(graph, directed=False, indices=[vertex])
    return _get_parcel_distance(vertex, dist, labels)


def _get_euclid_distance(vertex, coords, labels=None):
    """
    Gets Euclidean distance of `vertex` to all other vertices in `coords`.

    Parameters
    ----------
    vertex : int
        Index of vertex for which to calculate Euclidean distance
    coords : (N,3) array_like
        Coordinates of vertices on surface mesh
    labels : (N,) array_like, optional (default None)
        Labels indicating parcel to which each vertex belongs. If provided,
        distances will be averaged within M unique labels

    Returns
    -------
    dist : (N,) or (M,) np.ndarray
        Distance of `vertex` to all other vertices in `coords` (or to all
        unique parcels in `labels`, if provided)

    """
    dist = np.squeeze(cdist(coords[[vertex]], coords))
    return _get_parcel_distance(vertex, dist, labels)


def _get_parcel_distance(vertex, dist, labels=None):
    """
    Average `dist` within `labels`, if provided

    Parameters
    ----------
    vertex : int
        Index of vertex used to calculate `dist`
    dist : (N,) array_like
        Distance of `vertex` to all other vertices
    labels : (N,) array_like, optional (default None)
        Labels indicating parcel to which each vertex belongs. If provided,
        `dist` will be average within distinct labels.

    Returns
    -------
    dist : np.ndarray
        Distance from `vertex` to all vertices/parcels, cast to float32

    """

    if labels is not None:
        dist = ndimage.mean(input=np.delete(dist, vertex),
                            labels=np.delete(labels, vertex),
                            index=np.unique(labels))

    return np.atleast_2d(dist).astype(np.float32)



from .geo import cortex, subcortex, parcellate, volume
from .io import image2txt
from brainsmash.utils.dataio import load

__all__ = ['cortex', 'subcortex', 'parcellate', 'volume', 'image2txt', 'load']

# TODO "workbench" is now an a misleading name for this subpackage...



from pathlib import Path
import numpy as np

__all__ = ['check_extensions',
           'check_outfile',
           'check_pv',
           'check_deltas',
           'check_map',
           'check_distmat',
           'check_file_exists',
           'check_sampled',
           'is_string_like',
           'count_lines',
           'stripext']


def check_map(x):
    """
    Check that brain map conforms to expectations.

    Parameters
    ----------
    x : np.ndarray
        Brain map

    Returns
    -------
    None

    Raises
    ------
    TypeError : `x` is not a np.ndarray object
    ValueError : `x` is not one-dimensional

    """
    if not isinstance(x, np.ndarray):
        e = "Brain map must be array-like\n"
        e += "got type {}".format(type(x))
        raise TypeError(e)
    if x.ndim != 1:
        e = "Brain map must be one-dimensional\n"
        e += "got shape {}".format(x.shape)
        raise ValueError(e)


def check_extensions(filename, exts):
    """
    Test filename for a set of file extensions.

    Parameters
    ----------
    filename : str
        Path to file
    exts : List[str]
        List of allowed file extensions for `filename`

    Returns
    -------
    bool
        True if `filename`'s extensions is in `exts`

    Raises
    ------
    TypeError : `filename` is not string-like

    """
    if not is_string_like(filename):
        e = "Expected str, got {}".format(type(filename))
        e += "\nfilename: {}".format(filename)
        raise TypeError(e)
    ext = Path(filename).suffix
    return True if ext in exts else False


def check_distmat(D):
    """
    Check that a distance matrix conforms to expectations.

    Parameters
    ----------
    D : (N,N) np.ndarray
        Pairwise distance matrix

    Returns
    -------
    None

    Raises
    ------
    ValueError : `D` is not symmetric

    """
    if not np.allclose(D, D.T):
        raise ValueError("Distance matrix must be symmetric")


def check_sampled(D, index):
    """
    Check arguments provided to :class:`brainsmash.mapgen.sampled.Sampled`.

    Parameters
    ----------
    D : np.ndarray or np.memmap
        Pairwise distance matrix
    index : np.ndarray or np.memmap
        See :class:`brainsmash.mapgen.sampled.Sampled`

    Returns
    -------
    None

    Raises
    ------
    ValueError : Arguments do not have identical dimensions
    ValueError : `D` has not been sorted column-wise
    TypeError : rows of `D` or `index` are not sorted (ascending)

    """
    if not isinstance(D, np.ndarray) or not isinstance(index, np.ndarray):
        raise TypeError("'D' and 'index' must be array_like")
    if D.shape != index.shape:
        e = "`D` and `index` must have identical dimensions\n"
        e += "D.shape: {}".format(D.shape)
        e += "index.shape: {}".format(index.shape)
        raise ValueError(e)
    if isinstance(D, np.ndarray):
        if not np.all(D[:, 1:] >= D[:, :-1]):
            raise ValueError("Each row of `D` must be sorted (ascending)")
    else:  # just test the first row
        if not np.all(D[0, 1:] >= D[0, :-1]):
            raise ValueError("Each row of `D` must be sorted (ascending)")


def check_deltas(deltas):
    """
    Check input argument `deltas`.

    Parameters
    ----------
    deltas : np.ndarray or List[float]
        Proportions of neighbors to include for smoothing, in (0, 1]

    Returns
    -------
    None

    Raises
    ------
    TypeError : `deltas` is not a List or np.ndarray object
    ValueError : One or more elements of `deltas` lies outside (0,1]

    """
    if not isinstance(deltas, list) and not isinstance(deltas, np.ndarray):
        raise TypeError("Parameter `deltas` must be a list or ndarray")
    for d in deltas:
        if d <= 0 or d > 1:
            raise ValueError("Each element of `deltas` must lie in (0,1]")


def check_pv(pv):
    """
    Check input argument `pv`.

    Parameters
    ----------
    pv : int
        Percentile of the pairwise distance distribution at which to truncate
        during variogram fitting.

    Returns
    -------
    int

    Raises
    ------
    ValueError : `pv` lies outside range (0, 100]

    """
    try:
        pv = int(pv)
    except ValueError:
        raise ValueError("parameter 'pv' must be an integer in (0,100]")
    if pv <= 0 or pv > 100:
        raise ValueError("parameter 'pv' must be in (0,100]")
    return pv


def check_outfile(filename):
    """
    Warn if file exists and throw error if parent directory does not exist.

    Parameters
    ----------
    filename : filename
        File to be written

    Returns
    -------
    None

    Raises
    ------
    IOError : Parent directory of `filename` does not exist
    ValueError : directory was provided instead of file

    """
    if Path(filename).is_dir():
        raise ValueError("expected filename, got dir: {}".format(filename))
    if Path(filename).exists():
        print("WARNING: overwriting {}".format(filename))

    # Check that parent directory exists
    if not Path(filename).parent.exists():
        raise IOError("Output directory does not exist: {}".format(
            str(Path(filename).parent)))


def is_string_like(obj):
    """ Check whether `obj` behaves like a string. """
    try:
        obj + ''
    except (TypeError, ValueError):
        return False
    return True


def stripext(f):
    """
    Strip extension from a file.

    Parameters
    ----------
    f : filename
        Path to file with extension

    Returns
    -------
    objective : filename
        Path to file without extension

    """
    return str(Path(f).with_suffix(''))


def check_file_exists(f):
    """
    Check that file exists and has nonzero size.

    Parameters
    ----------
    f : filename

    Returns
    -------
    None

    Raises
    ------
    IOError : file does not exist or has zero size

    """
    if not Path(f).exists() or Path(f).stat().st_size == 0:
        raise IOError("{} was not successfully written to".format(f))


def count_lines(filename):
    """
    Count number of lines in a file.

    Parameters
    ----------
    filename : filename

    Returns
    -------
    int
        number of lines in file

    """
    with open(filename, 'rb') as f:
        lines = 0
        buf_size = 1024 * 1024
        read_f = f.raw.read
        buf = read_f(buf_size)
        while buf:
            lines += buf.count(b'\n')
            buf = read_f(buf_size)
        return lines



"""
Generate spatial autocorrelation-preserving surrogate maps from memory-mapped
arrays with random subsampling.
"""
from ..utils.dataio import dataio
from ..utils.checks import check_map, check_pv, check_deltas
from .kernels import check_kernel
from sklearn.linear_model import LinearRegression
from sklearn.utils.validation import check_random_state
import numpy as np
from joblib import Parallel, delayed

__all__ = ['Sampled']


class Sampled:
    """
    Sampling implementation of map generator.

    Parameters
    ----------
    x : filename or 1D np.ndarray
        Target brain map
    D : filename or (N,N) np.ndarray or np.memmap
        Pairwise distance matrix between elements of `x`. Each row of `D` should
        be sorted. Indices used to sort each row are passed to the `index`
        argument. See :func:`brainsmash.mapgen.memmap.txt2memmap` or the online
        documentation for more details (brainsmash.readthedocs.io)
    index : filename or (N,N) np.ndarray or np.memmap
        See above
    ns : int, default 500
        Take a subsample of `ns` rows from `D` when fitting variograms
    deltas : np.ndarray or List[float], default [0.3, 0.5, 0.7, 0.9]
        Proportions of neighbors to include for smoothing, in (0, 1]
    kernel : str, default 'exp'
        Kernel with which to smooth permuted maps
        - 'gaussian' : gaussian function
        - 'exp' : exponential decay function
        - 'invdist' : inverse distance
        - 'uniform' : uniform weights (distance independent)
    pv : int, default 70
        Percentile of the pairwise distance distribution (in `D`) at
        which to truncate during variogram fitting
    nh : int, default 25
        Number of uniformly spaced distances at which to compute variogram
    knn : int, default 1000
        Number of nearest regions to keep in the neighborhood of each region
    b : float or None, default None
        Gaussian kernel bandwidth for variogram smoothing. if None,
        three times the distance interval spacing is used.
    resample : bool, default False
        Resample surrogate map values from the target brain map
    verbose : bool, default False
        Print surrogate count each time new surrogate map created
    seed : None or int or np.random.RandomState instance (default None)
        Specify the seed for random number generation (or random state instance)
    n_jobs : int (default 1)
        Number of jobs to use for parallelizing creation of surrogate maps

    Notes
    -----
    Passing resample=True will preserve the distribution of values in the
    target map, at the expense of worsening simulated surrogate maps'
    variograms fits. This worsening will increase as the empirical map
    more strongly deviates from normality.

    Raises
    ------
    ValueError : `x` and `D` have inconsistent sizes

    """

    def __init__(self, x, D, index, ns=500, pv=70, nh=25, knn=1000, b=None,
                 deltas=np.arange(0.3, 1., 0.2), kernel='exp', resample=False,
                 verbose=False, seed=None, n_jobs=1):

        self._rs = check_random_state(seed)
        self._n_jobs = n_jobs

        self._verbose = verbose
        self.x = x
        n = self._x.size
        self.nmap = int(n)
        self.knn = knn
        self.D = D
        self.index = index
        self.resample = resample
        self.nh = int(nh)
        self.deltas = deltas
        self.ns = int(ns)
        self.b = b
        self.pv = pv
        self._ikn = np.arange(self._nmap)[:, None]

        # Store k nearest neighbors from distance and index matrices
        self.kernel = kernel  # Smoothing kernel selection
        self._dmax = np.percentile(self._D, self._pv)
        self.h = np.linspace(self._D.min(), self._dmax, self._nh)
        if not self._b:
            self.b = 3 * (self.h[1] - self.h[0])

        # Linear regression model
        self._lm = LinearRegression(fit_intercept=True)

    def __call__(self, n=1):
        """
        Randomly generate new surrogate map(s).

        Parameters
        ----------
        n : int, default 1
            Number of surrogate maps to randomly generate

        Returns
        -------
        (n,N) np.ndarray
            Randomly generated map(s) with matched spatial autocorrelation

        Notes
        -----
        Chooses a level of smoothing that produces a smoothed variogram which
        best approximates the true smoothed variogram. Selecting resample='True'
        preserves the map value distribution at the expense of worsening the
        surrogate maps' variogram fits.

        """

        rs = self._rs.randint(np.iinfo(np.int32).max, size=n)
        surrs = np.row_stack(
            Parallel(self._n_jobs)(
                delayed(self._call_method)(rs=i) for i in rs
            )
        )
        return np.asarray(surrs.squeeze())

    def _call_method(self, rs=None):
        """ Subfunction used by .__call__() for parallelization purposes """

        # Reset RandomState so parallel jobs yield different results
        self._rs = check_random_state(rs)

        # Randomly permute map
        x_perm = self.permute_map()

        # Randomly select subset of regions to use for variograms
        idx = self.sample()

        # Compute empirical variogram
        v = self.compute_variogram(self._x, idx)

        # Variogram ordinates; use nearest neighbors because local effect
        u = self._D[idx, :]
        uidx = np.where(u < self._dmax)

        # Smooth empirical variogram
        smvar, u0 = self.smooth_variogram(u[uidx], v[uidx], return_h=True)

        res = dict.fromkeys(self._deltas)

        for d in self._deltas:  # foreach neighborhood size

            k = int(d * self._knn)

            # Smooth the permuted map using k nearest neighbors to
            # reintroduce spatial autocorrelation
            sm_xperm = self.smooth_map(x=x_perm, k=k)

            # Calculate variogram values for the smoothed permuted map
            vperm = self.compute_variogram(sm_xperm, idx)

            # Calculate smoothed variogram of the smoothed permuted map
            smvar_perm = self.smooth_variogram(u[uidx], vperm[uidx])

            # Fit linear regression btwn smoothed variograms
            res[d] = self.regress(smvar_perm, smvar)

        alphas, betas, residuals = np.array(
            [res[d] for d in self._deltas], dtype=float).T

        # Select best-fit model and regression parameters
        iopt = np.argmin(residuals)
        dopt = self._deltas[iopt]
        self._dopt = dopt
        kopt = int(dopt * self._knn)
        aopt = alphas[iopt]
        bopt = betas[iopt]

        # Transform and smooth permuted map using best-fit parameters
        sm_xperm_best = self.smooth_map(x=x_perm, k=kopt)
        surr = (np.sqrt(np.abs(bopt)) * sm_xperm_best +
                np.sqrt(np.abs(aopt)) * self._rs.randn(self._nmap))

        if self._resample:  # resample values from empirical map
            sorted_map = np.sort(self._x)
            ii = np.argsort(surr)
            np.put(surr, ii, sorted_map)
        else:
            surr = surr - np.nanmean(surr) # De-mean

        if self._ismasked:
            return np.ma.masked_array(
                data=surr, mask=np.isnan(surr)).squeeze()
        return surr.squeeze()

    def compute_variogram(self, x, idx):
        """
        Compute variogram of `x` using pairs of regions indexed by `idx`.

        Parameters
        ----------
        x : (N,) np.ndarray
            Brain map
        idx : (ns,) np.ndarray[int]
            Indices of randomly sampled brain regions

        Returns
        -------
        v : (ns,ns) np.ndarray
            Variogram y-coordinates, i.e. 0.5 * (x_i - x_j) ^ 2, for i,j in idx

        """
        diff_ij = x[idx][:, None] - x[self._index[idx, :]]
        return 0.5 * np.square(diff_ij)

    def permute_map(self):
        """
        Return a random permutation of the target brain map.

        Returns
        -------
        (N,) np.ndarray
            Random permutation of target brain map

        """
        perm_idx = self._rs.permutation(self._nmap)
        if self._ismasked:
            mask_perm = self._x.mask[perm_idx]
            x_perm = self._x.data[perm_idx]
            return np.ma.masked_array(data=x_perm, mask=mask_perm)
        return self._x[perm_idx]

    def smooth_map(self, x, k):
        """
        Smooth `x` using `k` nearest neighboring regions.

        Parameters
        ----------
        x : (N,) np.ndarray
            Brain map
        k : float
            Number of nearest neighbors to include for smoothing

        Returns
        -------
        x_smooth : (N,) np.ndarray
            Smoothed brain map

        Notes
        -----
        Assumes `D` provided at runtime has been sorted.

        """
        jkn = self._index[:, :k]  # indices of k nearest neighbors
        xkn = x[jkn]  # values of k nearest neighbors
        dkn = self._D[:, :k]  # distances to k nearest neighbors
        weights = self._kernel(dkn)  # distance-weighted kernel
        # Kernel-weighted sum
        return (weights * xkn).sum(axis=1) / weights.sum(axis=1)

    def smooth_variogram(self, u, v, return_h=False):
        """
        Smooth a variogram.

        Parameters
        ----------
        u : (N,) np.ndarray
            Pairwise distances, ie variogram x-coordinates
        v : (N,) np.ndarray
            Squared differences, ie ariogram y-coordinates
        return_h : bool, default False
            Return distances at which smoothed variogram is computed

        Returns
        -------
        (nh,) np.ndarray
            Smoothed variogram samples
        (nh,) np.ndarray
            Distances at which smoothed variogram was computed (returned if
            `return_h` is True)

        Raises
        ------
        ValueError : `u` and `v` are not identically sized

        """
        if len(u) != len(v):
            raise ValueError("u and v must have same number of elements")

        # Subtract each element of h from each pairwise distance `u`.
        # Each row corresponds to a unique h.
        du = np.abs(u - self._h[:, None])
        w = np.exp(-np.square(2.68 * du / self._b) / 2)
        denom = np.nansum(w, axis=1)
        wv = w * v[None, :]
        num = np.nansum(wv, axis=1)
        output = num / denom
        if not return_h:
            return output
        return output, self._h

    def regress(self, x, y):
        """
        Linearly regress `x` onto `y`.

        Parameters
        ----------
        x : (N,) np.ndarray
            Independent variable
        y : (N,) np.ndarray
            Dependent variable

        Returns
        -------
        alpha : float
            Intercept term (offset parameter)
        beta : float
            Regression coefficient (scale parameter)
        res : float
            Sum of squared residuals

        """
        self._lm.fit(X=np.expand_dims(x, -1), y=y)
        beta = self._lm.coef_.item()
        alpha = self._lm.intercept_
        ypred = self._lm.predict(np.expand_dims(x, -1))
        res = np.sum(np.square(y-ypred))
        return alpha, beta, res

    def sample(self):
        """
        Randomly sample (without replacement) brain areas for variogram
        computation.

        Returns
        -------
        (self.ns,) np.ndarray
            Indices of randomly sampled areas

        """
        return self._rs.choice(
            a=self._nmap, size=self._ns, replace=False).astype(np.int32)

    @property
    def x(self):
        """ (N,) np.ndarray : brain map scalars """
        if self._ismasked:
            return np.ma.copy(self._x)
        return np.copy(self._x)

    @x.setter
    def x(self, x):
        self._ismasked = False
        x_ = dataio(x)
        check_map(x=x_)
        mask = np.isnan(x_)
        if mask.any():
            self._ismasked = True
            brain_map = np.ma.masked_array(data=x_, mask=mask)
        else:
            brain_map = x_
        self._x = brain_map

    @property
    def D(self):
        """ (N,N) np.memmap : Pairwise distance matrix """
        return np.copy(self._D)

    @D.setter
    def D(self, x):
        x_ = dataio(x)
        n = self._x.size
        if x_.shape[0] != n:
            raise ValueError(
                "D size along axis=0 must equal brain map size")
        self._D = x_[:, 1:self._knn + 1]  # prevent self-coupling

    @property
    def index(self):
        """ (N,N) np.memmap : indexes used to sort each row of dist. matrix """
        return np.copy(self._index)

    @index.setter
    def index(self, x):
        x_ = dataio(x)
        n = self._x.size
        if x_.shape[0] != n:
            raise ValueError(
                "index size along axis=0 must equal brain map size")
        self._index = x_[:, 1:self._knn+1].astype(np.int32)

    @property
    def nmap(self):
        """ int : length of brain map """
        return self._nmap

    @nmap.setter
    def nmap(self, x):
        self._nmap = int(x)

    @property
    def pv(self):
        """ int : percentile of pairwise distances at which to truncate """
        return self._pv

    @pv.setter
    def pv(self, x):
        pv = check_pv(x)
        self._pv = pv

    @property
    def deltas(self):
        """ np.ndarray or List[float] : proportions of nearest neighbors """
        return self._deltas

    @deltas.setter
    def deltas(self, x):
        check_deltas(deltas=x)
        self._deltas = x

    @property
    def nh(self):
        """ int : number of variogram distance intervals """
        return self._nh

    @nh.setter
    def nh(self, x):
        self._nh = x

    @property
    def kernel(self):
        """ Callable : smoothing kernel function

        Notes
        -----
        When setting kernel, use name of kernel as defined in ``config.py``.

        """
        return self._kernel

    @kernel.setter
    def kernel(self, x):
        kernel_callable = check_kernel(x)
        self._kernel = kernel_callable

    @property
    def resample(self):
        """ bool : whether to resample surrogate map values from target maps """
        return self._resample

    @resample.setter
    def resample(self, x):
        if not isinstance(x, bool):
            raise TypeError("expected bool, got {}".format(type(x)))
        self._resample = x

    @property
    def knn(self):
        """ int : number of nearest neighbors included in distance matrix """
        return self._knn

    @knn.setter
    def knn(self, x):
        if x > self._nmap:
            raise ValueError('knn must be less than len(X)')
        self._knn = int(x)

    @property
    def ns(self):
        """ int : number of randomly sampled regions used to construct map """
        return self._ns

    @ns.setter
    def ns(self, x):
        self._ns = int(x)

    @property
    def b(self):
        """ numeric : Gaussian kernel bandwidth """
        return self._b

    @b.setter
    def b(self, x):
        self._b = x

    @property
    def h(self):
        """ np.ndarray : distances at which variogram is evaluated """
        return self._h

    @h.setter
    def h(self, x):
        self._h = x
